"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

WA_GRAVITY = 9.81  # [m/s^2]
WA_EARTH_RADIUS = 6371000.0  # [m]
WA_PI = 3.141592653589793
